from enum import Enum
from typing import Sequence
from lark import Tree
from gersemi.command_invocation_dumpers.argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)
from gersemi.command_line_formatter import CommandLineFormatter
from gersemi.keyword_with_pairs_formatter import KeywordWithPairsFormatter
from gersemi.keywords import Keywords


class HintKind(Enum):
    CommandLine = "command_line"
    Pairs = "pairs"


def kind_to_function(kind: str) -> str:
    kind_enum = HintKind(kind) if kind in [e.value for e in HintKind] else None
    return {
        HintKind.Pairs: "_format_keyword_with_pairs",
        HintKind.CommandLine: "_format_command_line",
        None: "_default_format_values",
    }[kind_enum]


def create_specialized_dumper(
    raw_command_name: str, positional_arguments: Sequence[str], keywords: Keywords
):
    class Impl(
        ArgumentAwareCommandInvocationDumper,
        KeywordWithPairsFormatter,
        CommandLineFormatter,
    ):
        canonical_name = raw_command_name
        front_positional_arguments = positional_arguments
        options = keywords.options
        one_value_keywords = keywords.one_value_keywords
        multi_value_keywords = keywords.multi_value_keywords
        keyword_formatters = {
            hint.keyword: kind_to_function(hint.kind) for hint in keywords.hints
        }

        def custom_command(self, tree):
            _, command_name, arguments, *_ = tree.children
            return self.visit(Tree("command_invocation", [command_name, arguments]))

    return Impl



from contextlib import contextmanager
from functools import lru_cache
from lark import Tree
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.builtin_commands import BUILTIN_COMMANDS
from gersemi.command_invocation_dumpers.ctest_command_dumpers import (
    ctest_command_mapping,
)
from gersemi.command_invocation_dumpers.module_command_dumpers import (
    module_command_mapping,
)
from gersemi.command_invocation_dumpers.scripting_command_dumpers import (
    scripting_command_mapping,
)
from gersemi.command_invocation_dumpers.project_command_dumpers import (
    project_command_mapping,
)
from gersemi.command_invocation_dumpers.preserving_command_invocation_dumper import (
    PreservingCommandInvocationDumper,
)
from gersemi.command_invocation_dumpers.specialized_dumpers import (
    create_specialized_dumper,
)
from gersemi.keywords import Keywords


BUILTIN_COMMAND_DUMPERS = {
    **scripting_command_mapping,
    **project_command_mapping,
    **ctest_command_mapping,
    **module_command_mapping,
}


@lru_cache(maxsize=None)
def create_patch(patch, old_class):
    class Impl(patch, old_class):
        pass

    return Impl


def add_canonical_name(dumper_class, desired_canonical_name):
    class Impl(dumper_class):
        canonical_name = desired_canonical_name

    return Impl


class CommandInvocationDumper(
    PreservingCommandInvocationDumper, BaseCommandInvocationDumper
):
    @contextmanager
    def patched(self, patch):
        old_class = type(self)
        try:
            # pylint: disable=attribute-defined-outside-init
            self.__class__ = create_patch(patch, old_class)
            yield self
        finally:
            self.__class__ = old_class  # pylint: disable=invalid-class-object,

    def _get_patch(self, raw_command_name):
        command_name = raw_command_name.lower()
        if command_name in BUILTIN_COMMANDS:
            canonical_name = BUILTIN_COMMANDS[command_name]

            if command_name in BUILTIN_COMMAND_DUMPERS:
                return add_canonical_name(
                    BUILTIN_COMMAND_DUMPERS[command_name], canonical_name
                )

            return create_specialized_dumper(canonical_name, (), Keywords())

        if command_name in self.custom_command_definitions:
            canonical_name, arguments = self.custom_command_definitions[command_name]
            return create_specialized_dumper(canonical_name, *arguments)

        return None

    def command_invocation(self, tree):
        command_name, _ = tree.children
        patch = self._get_patch(command_name)
        if patch is None:
            return super().format_command(tree)
        with self.patched(patch):
            return self.format_command(tree)

    def custom_command(self, tree):
        _, command_name, arguments, *_ = tree.children
        if command_name.lower() in self.custom_command_definitions:
            return self.visit(
                Tree("command_invocation", [command_name.lower(), arguments])
            )
        return super().custom_command(tree)



import os


def is_not_comment_or_empty(s):
    return not (s.startswith("#") or len(s) == 0)


def get_builtin_commands():
    HERE = os.path.dirname(os.path.realpath(__file__))

    with open(os.path.join(HERE, "builtin_commands"), "r", encoding="utf-8") as f:
        return {
            item.lower(): item
            for item in filter(is_not_comment_or_empty, f.read().splitlines())
        }


BUILTIN_COMMANDS = get_builtin_commands()



from gersemi.configuration import Indent, ListExpansion
from gersemi.dumper import Dumper
from gersemi.parser import BARE_PARSER, PARSER
from gersemi.sanity_checker import check_code_equivalence


def noop(*_):
    pass


class Formatter:
    def __init__(  # pylint: disable=too-many-arguments
        self,
        sanity_checker,
        line_length: int,
        indent: Indent,
        custom_command_definitions,
        list_expansion: ListExpansion,
    ):
        self.sanity_checker = sanity_checker
        self.line_length = line_length
        self.indent_type = indent
        self.custom_command_definitions = custom_command_definitions
        self.list_expansion = list_expansion

    def format(self, code):
        dumper = Dumper(
            self.line_length,
            self.indent_type,
            self.custom_command_definitions,
            self.list_expansion,
        )
        result = dumper.visit(PARSER.parse(code))
        self.sanity_checker(BARE_PARSER, code, result)
        return result


class NullFormatter:
    def format(self, code):
        return code


def create_formatter(
    do_sanity_check,
    line_length,
    indent,
    custom_command_definitions,
    list_expansion,
):
    sanity_checker = check_code_equivalence if do_sanity_check else noop
    return Formatter(
        sanity_checker,
        line_length,
        indent,
        custom_command_definitions,
        list_expansion,
    )



from dataclasses import dataclass
from typing import Tuple, Union


@dataclass
class Hint:
    keyword: str
    kind: str


class AnyMatcher:
    def __eq__(self, other):
        return True

    def __hash__(self):
        return hash(str(self))


KeywordMatcher = Union[str, Tuple[str, Union[str, AnyMatcher]]]


@dataclass
class Keywords:
    options: Tuple[str, ...] = tuple()
    one_value_keywords: Tuple[KeywordMatcher, ...] = tuple()
    multi_value_keywords: Tuple[KeywordMatcher, ...] = tuple()
    hints: Tuple[Hint, ...] = tuple()



# pylint: disable=too-many-lines
from gersemi.command_line_formatter import CommandLineFormatter
from gersemi.keyword_with_pairs_formatter import KeywordWithPairsFormatter
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)
from .condition_syntax_command_invocation_dumper import (
    ConditionSyntaxCommandInvocationDumper,
)
from .multiple_signature_command_invocation_dumper import (
    MultipleSignatureCommandInvocationDumper,
)
from .two_word_keyword_isolator import TwoWordKeywordIsolator


_COMPARE_EQUAL = ("COMPARE", "EQUAL")
_COMPARE_GREATER = ("COMPARE", "GREATER")
_COMPARE_GREATER_EQUAL = ("COMPARE", "GREATER_EQUAL")
_COMPARE_LESS = ("COMPARE", "LESS")
_COMPARE_LESS_EQUAL = ("COMPARE", "LESS_EQUAL")
_COMPARE_NOTEQUAL = ("COMPARE", "NOTEQUAL")
_EVAL_CODE = ("EVAL", "CODE")
_EXTENSION_LAST_ONLY = ("EXTENSION", "LAST_ONLY")
_QUERY_WINDOWS_REGISTRY = ("QUERY", "WINDOWS_REGISTRY")
_REGEX_MATCH = ("REGEX", "MATCH")
_REGEX_MATCHALL = ("REGEX", "MATCHALL")
_REGEX_REPLACE = ("REGEX", "REPLACE")
_STEM_LAST_ONLY = ("STEM", "LAST_ONLY")


class Block(ArgumentAwareCommandInvocationDumper):
    inhibit_favour_expansion = True
    multi_value_keywords = ["SCOPE_FOR", "PROPAGATE"]


class CMakeHostSysteInformation(
    TwoWordKeywordIsolator, ArgumentAwareCommandInvocationDumper
):
    two_words_keywords = [_QUERY_WINDOWS_REGISTRY]
    one_value_keywords = [
        "RESULT",
        _QUERY_WINDOWS_REGISTRY,
        "VALUE_NAMES",
        "SUBKEYS",
        "VALUE",
        "VIEW",
        "SEPARATOR",
        "ERROR_VARIABLE",
    ]
    multi_value_keywords = ["QUERY"]


class CMakeLanguage(TwoWordKeywordIsolator, ArgumentAwareCommandInvocationDumper):
    two_words_keywords = [_EVAL_CODE]
    one_value_keywords = [
        "DIRECTORY",
        "ID",
        "ID_VAR",
        "GET_CALL_IDS",
        "GET_CALL",
        "SET_DEPENDENCY_PROVIDER",
        "GET_MESSAGE_LOG_LEVEL",
        "EXIT",
    ]
    multi_value_keywords = [
        "DEFER",
        "CALL",
        "CANCEL_CALL",
        "SUPPORTED_METHODS",
        _EVAL_CODE,
    ]


class CMakeMinimumRequired(ArgumentAwareCommandInvocationDumper):
    options = ["FATAL_ERROR"]
    one_value_keywords = ["VERSION"]


class CMakeParseArguments(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        "PARSE_ARGV": dict(
            front_positional_arguments=[
                "<N>",
                "<prefix>",
                "<options>",
                "<one_value_keywords>",
                "<multi_value_keywords>",
            ]
        ),
        None: dict(
            front_positional_arguments=[
                "<prefix>",
                "<options>",
                "<one_value_keywords>",
                "<multi_value_keywords>",
            ]
        ),
    }


class CMakePath(TwoWordKeywordIsolator, MultipleSignatureCommandInvocationDumper):
    two_words_keywords = [_EXTENSION_LAST_ONLY, _STEM_LAST_ONLY]
    customized_signatures = {
        # Decomposition
        "GET": {
            "front_positional_arguments": ["<path-var>"],
            "one_value_keywords": [
                "ROOT_NAME",
                "ROOT_DIRECTORY",
                "ROOT_PATH",
                "FILENAME",
                "EXTENSION",
                _EXTENSION_LAST_ONLY,
                "STEM",
                _STEM_LAST_ONLY,
                "RELATIVE_PART",
                "PARENT_PATH",
            ],
        },
        # Query
        "HAS_ROOT_NAME": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "HAS_ROOT_DIRECTORY": {
            "front_positional_arguments": ["<path-var>", "<out-var>"]
        },
        "HAS_ROOT_PATH": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "HAS_FILENAME": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "HAS_EXTENSION": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "HAS_STEM": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "HAS_RELATIVE_PATH": {
            "front_positional_arguments": ["<path-var>", "<out-var>"]
        },
        "HAS_PARENT_PATH": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "IS_ABSOLUTE": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "IS_RELATIVE": {"front_positional_arguments": ["<path-var>", "<out-var>"]},
        "IS_PREFIX": {
            "front_positional_arguments": ["<path-var>", "<out-var>"],
            "back_positional_arguments": ["<out-var>"],
            "options": ["NORMALIZE"],
        },
        "COMPARE": {
            "front_positional_arguments": ["<input1>", "<OP>", "<input2>", "<out-var>"]
        },
        # Modification
        "SET": {
            "front_positional_arguments": ["<path-var>"],
            "back_positional_arguments": ["<input>"],
            "options": ["NORMALIZE"],
        },
        "APPEND": {
            "front_positional_arguments": ["<path-var>"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        "APPEND_STRING": {
            "front_positional_arguments": ["<path-var>"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        "REMOVE_FILENAME": {
            "front_positional_arguments": ["<path-var>"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        "REPLACE_FILENAME": {
            "front_positional_arguments": ["<path-var>", "<input>"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        "REMOVE_EXTENSION": {
            "front_positional_arguments": ["<path-var>"],
            "options": ["LAST_ONLY"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        "REPLACE_EXTENSION": {
            "front_positional_arguments": ["<path-var>"],
            "options": ["LAST_ONLY"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        # Generation
        "NORMAL_PATH": {
            "front_positional_arguments": ["<path-var>"],
            "one_value_keywords": ["OUTPUT_VARIABLE"],
        },
        "RELATIVE_PATH": {
            "front_positional_arguments": ["<path-var>"],
            "one_value_keywords": ["BASE_DIRECTORY", "OUTPUT_VARIABLE"],
        },
        "ABSOLUTE_PATH": {
            "front_positional_arguments": ["<path-var>"],
            "options": ["NORMALIZE"],
            "one_value_keywords": ["BASE_DIRECTORY", "OUTPUT_VARIABLE"],
        },
        # Native Conversion
        "NATIVE_PATH": {
            "front_positional_arguments": ["<path-var>"],
            "back_positional_arguments": ["<out-var>"],
            "options": ["NORMALIZE"],
        },
        "CONVERT": {
            "front_positional_arguments": ["<input>"],
            "options": ["NORMALIZE"],
            "one_value_keywords": ["TO_CMAKE_PATH_LIST", "TO_NATIVE_PATH_LIST"],
        },
        # Hashing
        "HASH": {
            "front_positional_arguments": ["<path-var>", "<out-var>"],
        },
    }


class CMakePolicy(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        "VERSION": dict(front_positional_arguments=["<min>...<max>"]),
        "SET": dict(options=["OLD", "NEW"]),
        "GET": dict(back_positional_arguments=["<variable>"]),
        "PUSH": dict(),
        "POP": dict(),
    }


class ConfigureFile(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<input>", "<output>"]
    options = [
        "COPYONLY",
        "ESCAPE_QUOTES",
        "@ONLY",
        "NO_SOURCE_PERMISSIONS",
        "USE_SOURCE_PERMISSIONS",
    ]
    one_value_keywords = ["NEWLINE_STYLE"]
    multi_value_keywords = ["FILE_PERMISSIONS"]


class EndBlock(ArgumentAwareCommandInvocationDumper):
    pass


class EndForeach(ArgumentAwareCommandInvocationDumper):
    pass


class EndFunction(ArgumentAwareCommandInvocationDumper):
    pass


class EndMacro(ArgumentAwareCommandInvocationDumper):
    pass


class ExecuteProcess(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    options = [
        "OUTPUT_QUIET",
        "ERROR_QUIET",
        "OUTPUT_STRIP_TRAILING_WHITESPACE",
        "ERROR_STRIP_TRAILING_WHITESPACE",
        "ECHO_OUTPUT_VARIABLE",
        "ECHO_ERROR_VARIABLE",
    ]
    one_value_keywords = [
        "WORKING_DIRECTORY",
        "TIMEOUT",
        "RESULT_VARIABLE",
        "RESULTS_VARIABLE",
        "OUTPUT_VARIABLE",
        "ERROR_VARIABLE",
        "INPUT_FILE",
        "OUTPUT_FILE",
        "ERROR_FILE",
        "COMMAND_ECHO",
        "ENCODING",
        "COMMAND_ERROR_IS_FATAL",
    ]
    multi_value_keywords = ["COMMAND"]
    keyword_formatters = {"COMMAND": "_format_command_line"}


_GENERATE_OUTPUT = ("GENERATE", "OUTPUT")


class File(TwoWordKeywordIsolator, MultipleSignatureCommandInvocationDumper):
    two_words_keywords = [_GENERATE_OUTPUT]
    customized_signatures = {
        # Reading
        "READ": dict(
            front_positional_arguments=["<filename>", "<variable>"],
            options=["HEX"],
            one_value_keywords=["OFFSET", "LIMIT"],
        ),
        "STRINGS": dict(
            front_positional_arguments=["<filename>", "<variable>"],
            options=["NEWLINE_CONSUME", "NO_HEX_CONVERSION"],
            one_value_keywords=[
                "LENGTH_MAXIMUM",
                "LENGTH_MINIMUM",
                "LIMIT_COUNT",
                "LIMIT_INPUT",
                "LIMIT_OUTPUT",
                "REGEX",
                "ENCODING",
            ],
        ),
        "MD5": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA1": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA224": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA256": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA384": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA512": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_224": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_256": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_384": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_512": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "TIMESTAMP": dict(
            front_positional_arguments=["<filename>", "<variable>", "<format>"],
            options=["UTC"],
        ),
        "GET_RUNTIME_DEPENDENCIES": dict(
            one_value_keywords=[
                "RESOLVED_DEPENDENCIES_VAR",
                "UNRESOLVED_DEPENDENCIES_VAR",
                "CONFLICTING_DEPENDENCIES_PREFIX",
                "BUNDLE_EXECUTABLE",
            ],
            multi_value_keywords=[
                "EXECUTABLES",
                "LIBRARIES",
                "MODULES",
                "DIRECTORIES",
                "PRE_INCLUDE_REGEXES",
                "PRE_EXCLUDE_REGEXES",
                "POST_INCLUDE_REGEXES",
                "POST_EXCLUDE_REGEXES",
                "POST_INCLUDE_FILES",
                "POST_EXCLUDE_FILES",
            ],
        ),
        # Writing
        "WRITE": dict(front_positional_arguments=["<filename>"]),
        "APPEND": dict(front_positional_arguments=["<filename>"]),
        _GENERATE_OUTPUT: dict(
            front_positional_arguments=["output-file"],
            options=["NO_SOURCE_PERMISSIONS", "USE_SOURCE_PERMISSIONS"],
            one_value_keywords=[
                "INPUT",
                "CONTENT",
                "CONDITION",
                "TARGET",
                "NEWLINE_STYLE",
            ],
            multi_value_keywords=["FILE_PERMISSIONS"],
        ),
        "CONFIGURE": dict(
            options=["ESCAPE_QUOTES", "@ONLY"],
            one_value_keywords=["OUTPUT", "CONTENT", "NEWLINE_STYLE"],
        ),
        # Filesystem
        "GLOB": dict(
            front_positional_arguments=["<variable>"],
            options=["CONFIGURE_DEPENDS"],
            one_value_keywords=["GLOB", "LIST_DIRECTORIES", "RELATIVE"],
        ),
        "GLOB_RECURSE": dict(
            front_positional_arguments=["<variable>"],
            options=["CONFIGURE_DEPENDS", "FOLLOW_SYMLINKS"],
            one_value_keywords=["GLOB_RECURSE", "LIST_DIRECTORIES", "RELATIVE"],
        ),
        "RENAME": dict(
            front_positional_arguments=["<oldname>", "<newname>"],
            options=["NO_REPLACE"],
            one_value_keywords=["RESULT"],
        ),
        "COPY": dict(
            options=[
                "NO_SOURCE_PERMISSIONS",
                "USE_SOURCE_PERMISSIONS",
                "FOLLOW_SYMLINK_CHAIN",
                "FILES_MATCHING",
                "EXCLUDE",
            ],
            one_value_keywords=["DESTINATION", "PATTERN", "REGEX"],
            multi_value_keywords=[
                "COPY",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
                "PERMISSIONS",
            ],
        ),
        "COPY_FILE": dict(
            front_positional_arguments=["<oldname>", "<newname>"],
            options=["ONLY_IF_DIFFERENT"],
            one_value_keywords=["RESULT"],
        ),
        "INSTALL": dict(
            options=[
                "NO_SOURCE_PERMISSIONS",
                "USE_SOURCE_PERMISSIONS",
                "FOLLOW_SYMLINK_CHAIN",
                "FILES_MATCHING",
                "EXCLUDE",
            ],
            one_value_keywords=["DESTINATION", "PATTERN", "REGEX"],
            multi_value_keywords=[
                "INSTALL",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
                "PERMISSIONS",
            ],
        ),
        "SIZE": dict(front_positional_arguments=["<filename>", "<variable>"]),
        "READ_SYMLINK": dict(front_positional_arguments=["<linkname>", "<variable>"]),
        "CREATE_LINK": dict(
            front_positional_arguments=["<original>", "<linkname>"],
            options=["COPY_ON_ERROR", "SYMBOLIC"],
            one_value_keywords=["RESULT"],
            multi_value_keywords=["CREATE_LINK"],
        ),
        "CHMOD": dict(
            multi_value_keywords=[
                "PERMISSIONS",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
            ]
        ),
        "CHMOD_RECURSE": dict(
            multi_value_keywords=[
                "PERMISSIONS",
                "FILE_PERMISSIONS",
                "DIRECTORY_PERMISSIONS",
            ]
        ),
        # Path Conversion
        "REAL_PATH": dict(
            front_positional_arguments=["<path>", "<out-var>"],
            options=["EXPAND_TILDE"],
            one_value_keywords=["BASE_DIRECTORY"],
        ),
        "RELATIVE_PATH": dict(
            front_positional_arguments=["<variable>", "<directory>", "<file>"]
        ),
        "TO_CMAKE_PATH": dict(front_positional_arguments=["<path>", "<variable>"]),
        "TO_NATIVE_PATH": dict(front_positional_arguments=["<path>", "<variable>"]),
        # Transfer
        "DOWNLOAD": dict(
            front_positional_arguments=["<url>", "<file>"],
            options=["SHOW_PROGRESS"],
            one_value_keywords=[
                "INACTIVITY_TIMEOUT",
                "LOG",
                "STATUS",
                "TIMEOUT",
                "USERPWD",
                "HTTPHEADER",
                "NETRC",
                "NETRC_FILE",
                "EXPECTED_HASH",
                "EXPECTED_MD5",
                "TLS_VERIFY",
                "TLS_CAINFO",
                "RANGE_START",
                "RANGE_END",
                "TLS_VERSION",
            ],
            multi_value_keywords=["DOWNLOAD"],
        ),
        "UPLOAD": dict(
            front_positional_arguments=["<file>", "<url>"],
            options=["SHOW_PROGRESS"],
            one_value_keywords=[
                "INACTIVITY_TIMEOUT",
                "LOG",
                "STATUS",
                "TIMEOUT",
                "USERPWD",
                "HTTPHEADER",
                "NETRC",
                "NETRC_FILE",
                "TLS_VERIFY",
                "TLS_CAINFO",
                "TLS_VERSION",
            ],
            multi_value_keywords=["UPLOAD"],
        ),
        # Locking
        "LOCK": dict(
            front_positional_arguments=["<path>"],
            options=["DIRECTORY", "RELEASE"],
            one_value_keywords=["LOCK", "GUARD", "RESULT_VARIABLE", "TIMEOUT"],
        ),
        # Archiving
        "ARCHIVE_CREATE": dict(
            options=["VERBOSE"],
            one_value_keywords=[
                "OUTPUT",
                "FORMAT",
                "COMPRESSION",
                "COMPRESSION_LEVEL",
                "MTIME",
            ],
            multi_value_keywords=["PATHS"],
        ),
        "ARCHIVE_EXTRACT": dict(
            options=["LIST_ONLY", "VERBOSE"],
            one_value_keywords=["INPUT", "DESTINATION"],
            multi_value_keywords=["PATTERNS"],
        ),
    }


class FindFile(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<VAR>", "name"]
    options = [
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
        "REQUIRED",
        "NO_CMAKE_INSTALL_PREFIX",
    ]
    one_value_keywords = ["DOC", "ENV", "VALIDATOR"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class FindLibrary(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<VAR>", "name"]
    options = [
        "NAMES_PER_DIR",
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
        "REQUIRED",
        "NO_CMAKE_INSTALL_PREFIX",
    ]
    one_value_keywords = ["DOC", "ENV", "VALIDATOR"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class FindPackage(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<PackageName>", "<version>"]
    options = [
        "EXACT",
        "QUIET",
        "MODULE",
        "CONFIG",
        "NO_MODULE",
        "NO_POLICY_SCOPE",
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_PACKAGE_REGISTRY",
        "NO_CMAKE_BUILDS_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "NO_CMAKE_SYSTEM_PACKAGE_REGISTRY",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_INSTALL_PREFIX",
        "GLOBAL",
    ]
    multi_value_keywords = [
        "COMPONENTS",
        "OPTIONAL_COMPONENTS",
        "NAMES",
        "CONFIGS",
        "HINTS",
        "PATHS",
        "PATH_SUFFIXES",
        "REQUIRED",
    ]


class FindPath(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<VAR>", "name"]
    options = [
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
        "REQUIRED",
        "NO_CMAKE_INSTALL_PREFIX",
    ]
    one_value_keywords = ["DOC", "ENV", "VALIDATOR"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class FindProgram(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<VAR>", "name"]
    options = [
        "NAMES_PER_DIR",
        "NO_DEFAULT_PATH",
        "NO_PACKAGE_ROOT_PATH",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "NO_SYSTEM_ENVIRONMENT_PATH",
        "NO_CMAKE_SYSTEM_PATH",
        "CMAKE_FIND_ROOT_PATH_BOTH",
        "ONLY_CMAKE_FIND_ROOT_PATH",
        "NO_CMAKE_FIND_ROOT_PATH",
        "REQUIRED",
        "NO_CMAKE_INSTALL_PREFIX",
    ]
    one_value_keywords = ["DOC", "ENV", "VALIDATOR"]
    multi_value_keywords = ["NAMES", "HINTS", "PATHS", "PATH_SUFFIXES"]


class Foreach(ArgumentAwareCommandInvocationDumper):
    inhibit_favour_expansion = True
    front_positional_arguments = ["<loop_var>"]
    options = ["IN"]
    multi_value_keywords = ["RANGE", "LISTS", "ITEMS", "ZIP_LISTS"]


class Function(ArgumentAwareCommandInvocationDumper):
    inhibit_favour_expansion = True
    front_positional_arguments = ["<name>"]


class GetCMakeProperty(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<var>", "<property>"]


class GetDirectoryProperty(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>"]
    one_value_keywords = ["DIRECTORY", "DEFINITION"]


class GetFilenameComponent(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = [
        "<var>",
        "<FileName>",
        "<mode>",
    ]
    options = [
        "DIRECTORY",
        "NAME",
        "EXT",
        "NAME_WE",
        "LAST_EXT",
        "NAME_WLE",
        "PATH",
        "PROGRAM",
        "CACHE",
    ]
    one_value_keywords = ["BASE_DIR", "PROGRAM_ARGS"]


class GetProperty(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>"]
    options = ["GLOBAL", "VARIABLE", "SET", "DEFINED", "BRIEF_DOCS", "FULL_DOCS"]
    one_value_keywords = [
        "TARGET",
        "INSTALL",
        "TEST",
        "CACHE",
        "PROPERTY",
        "TARGET_DIRECTORY",
        "SOURCE",
    ]
    multi_value_keywords = ["DIRECTORY"]


class GetSourceFileProperty(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DIRECTORY", "TARGET_DIRECTORY"]


class Include(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<file|module>"]
    options = ["OPTIONAL", "NO_POLICY_SCOPE"]
    one_value_keywords = ["RESULT_VARIABLE"]


class List(MultipleSignatureCommandInvocationDumper):
    customized_signatures = {
        # Reading
        "LENGTH": dict(front_positional_arguments=["<list>", "<output variable>"]),
        "GET": dict(
            front_positional_arguments=["<list>"],
            back_positional_arguments=["<output variable>"],
        ),
        "JOIN": dict(
            front_positional_arguments=["<list>", "<glue", "<output variable>"]
        ),
        "SUBLIST": dict(
            front_positional_arguments=[
                "<list>",
                "<begin>",
                "<length>",
                "<out-var>",
            ],
        ),
        # Search
        "FIND": dict(
            front_positional_arguments=["<list>", "<value>", "<output variable>"]
        ),
        # Modification
        "APPEND": dict(front_positional_arguments=["<list>"]),
        "FILTER": dict(
            options=["INCLUDE", "EXCLUDE"], one_value_keywords=["FILTER", "REGEX"]
        ),
        "INSERT": dict(front_positional_arguments=["<list>", "<element_index>"]),
        "POP_BACK": dict(front_positional_arguments=["<list>"]),
        "POP_FRONT": dict(front_positional_arguments=["<list>"]),
        "PREPEND": dict(front_positional_arguments=["<list>"]),
        "REMOVE_ITEM": dict(front_positional_arguments=["<list>"]),
        "REMOVE_AT": dict(front_positional_arguments=["<list>"]),
        "REMOVE_DUPLICATES": dict(front_positional_arguments=["<list>"]),
        "TRANSFORM": dict(
            front_positional_arguments=["<list>"],
            one_value_keywords=["OUTPUT_VARIABLE", "TRANSFORM"],
            multi_value_keywords=[
                "APPEND",
                "PREPEND",
                "TOLOWER",
                "TOUPPER",
                "STRIP",
                "GENEX_STRIP",
                "REPLACE",
            ],
        ),
        # Ordering
        "REVERSE": dict(front_positional_arguments=["<list>"]),
        "SORT": dict(one_value_keywords=["SORT", "COMPARE", "CASE", "ORDER"]),
    }


class Macro(ArgumentAwareCommandInvocationDumper):
    inhibit_favour_expansion = True
    front_positional_arguments = ["<name>"]


class MarkAsAdvanced(ArgumentAwareCommandInvocationDumper):
    options = ["CLEAR", "FORCE"]


class Math(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["EXPR", "<variable>", "<expression>"]
    one_value_keywords = ["OUTPUT_FORMAT"]


class Message(ArgumentAwareCommandInvocationDumper):
    options = [
        "FATAL_ERROR",
        "SEND_ERROR",
        "WARNING",
        "AUTHOR_WARNING",
        "DEPRECATION",
        "NOTICE",
        "STATUS",
        "VERBOSE",
        "DEBUG",
        "TRACE",
        "CHECK_START",
        "CHECK_PASS",
        "CHECK_FAIL",
        "CONFIGURE_LOG",
    ]


class Option(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>", "<help_text>", "<value>"]


class Return(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["PROPAGATE"]


class SeparateArguments(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>", "<mode>"]
    options = ["PROGRAM", "SEPARATE_ARGS"]


class SetProperty(ArgumentAwareCommandInvocationDumper):
    options = ["GLOBAL", "APPEND", "APPEND_STRING"]
    multi_value_keywords = [
        "TARGET",
        "SOURCE",
        "INSTALL",
        "TEST",
        "CACHE",
        "PROPERTY",
        "TARGET_DIRECTORIES",
        "DIRECTORY",
    ]
    keyword_formatters = {"PROPERTY": "_format_property"}

    def _format_property(self, args):
        result = self._try_to_format_into_single_line(
            args, separator=" ", prefix="(", postfix=")"
        )
        if result is not None:
            return result

        name, *rest = args
        with self.indented():
            formatted_rest = "\n".join(map(self.visit, rest))
        return f"{self.visit(name)}\n{formatted_rest}"


class Set(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>"]
    options = ["PARENT_SCOPE", "FORCE"]
    one_value_keywords = ["CACHE"]


class String(TwoWordKeywordIsolator, MultipleSignatureCommandInvocationDumper):
    two_words_keywords = [
        _REGEX_MATCH,
        _REGEX_MATCHALL,
        _REGEX_REPLACE,
        _COMPARE_LESS,
        _COMPARE_GREATER,
        _COMPARE_EQUAL,
        _COMPARE_NOTEQUAL,
        _COMPARE_LESS_EQUAL,
        _COMPARE_GREATER_EQUAL,
    ]
    customized_signatures = {
        # Search and Replace
        "FIND": dict(
            front_positional_arguments=["<string>", "<substring>", "<output variable>"],
            options=["REVERSE"],
        ),
        "REPLACE": dict(
            front_positional_arguments=[
                "<match_string>",
                "<replace_string>",
                "<output_variable>",
            ]
        ),
        # Regular Expressions
        _REGEX_MATCH: dict(
            front_positional_arguments=["<regular_expression>", "<output_variable>"]
        ),
        _REGEX_MATCHALL: dict(
            front_positional_arguments=["<regular_expression>", "<output_variable>"]
        ),
        _REGEX_REPLACE: dict(
            front_positional_arguments=[
                "<regular_expression>",
                "<replacement_expression>",
                "<output_variable>",
            ]
        ),
        # Manipulation
        "APPEND": dict(front_positional_arguments=["<string_variable>"]),
        "PREPEND": dict(front_positional_arguments=["<string_variable>"]),
        "CONCAT": dict(front_positional_arguments=["<output_variable>"]),
        "JOIN": dict(front_positional_arguments=["<glue>", "<output_variable>"]),
        "TOLOWER": dict(front_positional_arguments=["<string>", "<output_variable>"]),
        "TOUPPER": dict(front_positional_arguments=["<string>", "<output_variable>"]),
        "LENGTH": dict(front_positional_arguments=["<string>", "<output_variable>"]),
        "SUBSTRING": dict(
            front_positional_arguments=[
                "<string>",
                "<begin>",
                "<length>",
                "<output_variable>",
            ]
        ),
        "STRIP": dict(front_positional_arguments=["<string>", "<output_variable>"]),
        "GENEX_STRIP": dict(
            front_positional_arguments=["<string>", "<output_variable>"]
        ),
        "REPEAT": dict(
            front_positional_arguments=["<string>", "<count>", "<output_variable>"]
        ),
        # Comparison
        _COMPARE_LESS: dict(
            front_positional_arguments=["<string1>", "<string2>", "<output_variable>"]
        ),
        _COMPARE_GREATER: dict(
            front_positional_arguments=["<string1>", "<string2>", "<output_variable>"]
        ),
        _COMPARE_EQUAL: dict(
            front_positional_arguments=["<string1>", "<string2>", "<output_variable>"]
        ),
        _COMPARE_NOTEQUAL: dict(
            front_positional_arguments=["<string1>", "<string2>", "<output_variable>"]
        ),
        _COMPARE_LESS_EQUAL: dict(
            front_positional_arguments=["<string1>", "<string2>", "<output_variable>"]
        ),
        _COMPARE_GREATER_EQUAL: dict(
            front_positional_arguments=["<string1>", "<string2>", "<output_variable>"]
        ),
        # Hashing
        "MD5": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA1": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA224": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA256": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA384": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA512": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_224": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_256": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_384": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        "SHA3_512": dict(front_positional_arguments=["<output_variable>", "<input>"]),
        # Generation
        "ASCII": dict(
            back_positional_arguments=["<output_variable>"],
        ),
        "HEX": dict(front_positional_arguments=["<string>", "<output_variable>"]),
        "CONFIGURE": dict(
            front_positional_arguments=["<string>", "<output_variable>"],
            options=["@ONLY", "ESCAPE_QUOTES"],
        ),
        "MAKE_C_IDENTIFIER": dict(
            front_positional_arguments=["<string>", "<output_variable>"]
        ),
        "RANDOM": dict(one_value_keywords=["LENGTH", "ALPHABET", "RANDOM_SEED"]),
        "TIMESTAMP": dict(
            front_positional_arguments=["<filename>", "<variable>", "<format>"],
            options=["UTC"],
        ),
        "UUID": dict(
            front_positional_arguments=["<output_variable>"],
            options=["UPPER"],
            one_value_keywords=["NAMESPACE", "NAME", "TYPE"],
        ),
        # JSON
        "JSON": dict(
            front_positional_arguments=["<out-var>"],
            one_value_keywords=[
                "ERROR_VARIABLE",
                "GET",
                "TYPE",
                "MEMBER",
                "LENGTH",
                "REMOVE",
                "SET",
            ],
            multi_value_keywords=["EQUAL"],
        ),
    }


class SetDirectoryProperties(
    KeywordWithPairsFormatter, ArgumentAwareCommandInvocationDumper
):
    multi_value_keywords = ["PROPERTIES"]
    keyword_formatters = {"PROPERTIES": "_format_keyword_with_pairs"}


class Unset(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<variable>"]
    options = ["CACHE", "PARENT_SCOPE"]


scripting_command_mapping = {
    "block": Block,
    "cmake_host_system_information": CMakeHostSysteInformation,
    "cmake_language": CMakeLanguage,
    "cmake_minimum_required": CMakeMinimumRequired,
    "cmake_parse_arguments": CMakeParseArguments,
    "cmake_path": CMakePath,
    "cmake_policy": CMakePolicy,
    "configure_file": ConfigureFile,
    "elseif": ConditionSyntaxCommandInvocationDumper,
    "else": ConditionSyntaxCommandInvocationDumper,
    "endif": ConditionSyntaxCommandInvocationDumper,
    "endblock": EndBlock,
    "endforeach": EndForeach,
    "endfunction": EndFunction,
    "endmacro": EndMacro,
    "endwhile": ConditionSyntaxCommandInvocationDumper,
    "execute_process": ExecuteProcess,
    "file": File,
    "find_file": FindFile,
    "find_library": FindLibrary,
    "find_package": FindPackage,
    "find_path": FindPath,
    "find_program": FindProgram,
    "foreach": Foreach,
    "function": Function,
    "get_cmake_property": GetCMakeProperty,
    "get_directory_property": GetDirectoryProperty,
    "get_filename_component": GetFilenameComponent,
    "get_property": GetProperty,
    "get_source_file_property": GetSourceFileProperty,
    "if": ConditionSyntaxCommandInvocationDumper,
    "include": Include,
    "list": List,
    "macro": Macro,
    "mark_as_advanced": MarkAsAdvanced,
    "math": Math,
    "message": Message,
    "option": Option,
    "return": Return,
    "separate_arguments": SeparateArguments,
    "set_directory_properties": SetDirectoryProperties,
    "set_property": SetProperty,
    "set": Set,
    "string": String,
    "unset": Unset,
    "while": ConditionSyntaxCommandInvocationDumper,
}



from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class CTestBuild(ArgumentAwareCommandInvocationDumper):
    options = ["APPEND", "QUIET"]
    one_value_keywords = [
        "BUILD",
        "CONFIGURATION",
        "FLAGS",
        "PROJECT_NAME",
        "TARGET",
        "NUMBER_ERRORS",
        "NUMBER_WARNINGS",
        "RETURN_VALUE",
        "CAPTURE_CMAKE_ERROR",
        "PARALLEL_LEVEL",
    ]


class CTestConfigure(ArgumentAwareCommandInvocationDumper):
    options = ["APPEND", "QUIET"]
    one_value_keywords = [
        "BUILD",
        "SOURCE",
        "OPTIONS",
        "RETURN_VALUE",
        "CAPTURE_CMAKE_ERROR",
    ]


class CTestCoverage(ArgumentAwareCommandInvocationDumper):
    options = ["APPEND", "QUIET"]
    one_value_keywords = ["BUILD", "RETURN_VALUE", "CAPTURE_CMAKE_ERROR"]
    multi_value_keywords = ["LABELS"]


class CTestMemcheck(ArgumentAwareCommandInvocationDumper):
    options = ["APPEND", "QUIET"]
    one_value_keywords = [
        "BUILD",
        "START",
        "END",
        "STRIDE",
        "EXCLUDE",
        "INCLUDE",
        "EXCLUDE_LABEL",
        "INCLUDE_LABEL",
        "EXCLUDE_FIXTURE",
        "EXCLUDE_FIXTURE_SETUP",
        "EXCLUDE_FIXTURE_CLEANUP",
        "PARALLEL_LEVEL",
        "TEST_LOAD",
        "SCHEDULE_RANDOM",
        "STOP_TIME",
        "RETURN_VALUE",
        "DEFECT_COUNT",
    ]


class CTestRunScript(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["RETURN_VALUE"]


class CTestStart(ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["<model>", "<source>", "<binary>"]
    options = ["APPEND", "QUIET"]
    one_value_keywords = ["GROUP"]


class CTestSubmit(ArgumentAwareCommandInvocationDumper):
    options = ["QUIET"]
    one_value_keywords = [
        "SUBMIT_URL",
        "BUILD_ID",
        "HTTPHEADER",
        "RETRY_COUNT",
        "RETRY_DELAY",
        "RETURN_VALUE",
        "CAPTURE_CMAKE_ERROR",
        "CDASH_UPLOAD",
        "CDASH_UPLOAD_TYPE",
    ]
    multi_value_keywords = ["PARTS", "FILES"]


class CTestTest(ArgumentAwareCommandInvocationDumper):
    options = ["APPEND", "QUIET"]
    one_value_keywords = [
        "BUILD",
        "START",
        "END",
        "STRIDE",
        "EXCLUDE",
        "INCLUDE",
        "EXCLUDE_LABEL",
        "INCLUDE_LABEL",
        "EXCLUDE_FIXTURE",
        "EXCLUDE_FIXTURE_SETUP",
        "EXCLUDE_FIXTURE_CLEANUP",
        "PARALLEL_LEVEL",
        "RESOURCE_SPEC_FILE",
        "TEST_LOAD",
        "SCHEDULE_RANDOM",
        "STOP_TIME",
        "RETURN_VALUE",
        "CAPTURE_CMAKE_ERROR",
        "INCLUDE_FROM_FILE",
        "EXCLUDE_FROM_FILE",
    ]


class CTestUpdate(ArgumentAwareCommandInvocationDumper):
    options = ["QUIET"]
    one_value_keywords = ["SOURCE", "RETURN_VALUE", "CAPTURE_CMAKE_ERROR"]


class CTestUpload(ArgumentAwareCommandInvocationDumper):
    options = ["QUIET"]
    one_value_keywords = ["CAPTURE_CMAKE_ERROR"]
    multi_value_keywords = ["FILES"]


ctest_command_mapping = {
    "ctest_build": CTestBuild,
    "ctest_configure": CTestConfigure,
    "ctest_coverage": CTestCoverage,
    "ctest_memcheck": CTestMemcheck,
    "ctest_run_script": CTestRunScript,
    "ctest_start": CTestStart,
    "ctest_submit": CTestSubmit,
    "ctest_test": CTestTest,
    "ctest_update": CTestUpdate,
    "ctest_upload": CTestUpload,
}



import pytest
from gersemi.exceptions import ASTMismatch
from gersemi.sanity_checker import check_code_equivalence
from .tests_generator import generate_input_output_tests


def preprocess(text):
    return (
        text.replace("⟶", "\\⟶")
        .replace("·", "\\·")
        .replace("\t", "⟶")
        .replace(" ", "·")
    )


def test_formatter(formatter_creator, case):
    p = preprocess
    formatter = formatter_creator(case.config)
    assert p(formatter.format(case.given)) == p(case.expected)


def test_formatter_idempotence(formatter_creator, case):
    p = preprocess
    formatter = formatter_creator(case.config)
    formatted_once = formatter.format(case.given)
    formatted_twice = formatter.format(formatted_once)
    assert p(formatted_once) == p(formatted_twice)


def test_abstract_syntax_tree_equivalence(parser, parser_with_simple_grammar, case):
    for p in [parser, parser_with_simple_grammar]:
        try:
            check_code_equivalence(p, case.given, case.expected)
        except ASTMismatch:
            pytest.fail("ASTs mismatch")
            raise


pytest_generate_tests = generate_input_output_tests(
    where="formatter", input_extension=".in.cmake", output_extension=".out.cmake"
)



from enum import Enum


class Mode(Enum):
    ForwardToStdout = 0
    RewriteInPlace = 1
    CheckFormatting = 2
    ShowDiff = 3


def get_mode(args) -> Mode:
    if args.show_diff:
        return Mode.ShowDiff
    if args.check_formatting:
        return Mode.CheckFormatting
    if args.in_place:
        return Mode.RewriteInPlace
    return Mode.ForwardToStdout



from contextlib import contextmanager
from dataclasses import asdict, dataclass, field, fields
from enum import Enum
from hashlib import sha1
from itertools import chain
import multiprocessing
import os
from pathlib import Path
import sys
import textwrap
from typing import Iterable, Optional, Union
import yaml


def number_of_workers():
    result = multiprocessing.cpu_count()
    if sys.platform == "win32":
        # https://bugs.python.org/issue26903
        # https://github.com/python/cpython/issues/89240
        return min(result, 60)
    return result


def doc(text: str) -> str:
    return " ".join(textwrap.dedent(text).splitlines()).strip()


class EnumWithMetadata(Enum):
    def __new__(cls, data):
        self = object.__new__(cls)
        self._value_ = data["value"]
        self.description = doc(data["description"])
        self.title = data["title"]
        return self


class Tabs(EnumWithMetadata):
    Tabs = dict(
        value="tabs",
        description="Use tabs to indent the code.",
        title="Tab indent",
    )


Spaces = int
Indent = Union[Spaces, Tabs]


def indent_type(thing) -> Indent:
    if thing == Tabs.Tabs.value:
        return Tabs.Tabs
    return int(thing)


class ListExpansion(EnumWithMetadata):
    FavourInlining = dict(
        value="favour-inlining",
        description="""
    With "favour-inlining" the list of entities will be formatted in such way that sublists
    might still be formatted into single line as long as it's possible or as long as it doesn't
    break the "more than four standalone arguments" heuristic that's mostly focused on commands
    like `set` or `list(APPEND)`.
        """,
        title="Favour inlining",
    )
    FavourExpansion = dict(
        value="favour-expansion",
        description="""
    With "favour-expansion" the list of entities will be formatted in such way that sublists
    will be completely expanded once expansion becomes necessary at all.
        """,
        title="Favour expansion",
    )


@dataclass
class Configuration:  # pylint: disable=too-many-instance-attributes
    """
    By default configuration is loaded from YAML formatted .gersemirc file if it's available.
    This file should be placed in one of the common parent directories of source files.
    Arguments from command line can be used to override parts of that configuration or
    supply them in absence of configuration file.
    """

    line_length: int = field(
        default=80,
        metadata=dict(
            title="Line length",
            description="Maximum line length in characters.",
        ),
    )

    indent: Indent = field(
        default=4,
        metadata=dict(
            title="Indent",
            description="Number of spaces used to indent or 'tabs' for indenting with tabs",
        ),
    )

    unsafe: bool = field(
        default=False,
        metadata=dict(
            title="Unsafe",
            description="Skip default sanity checks.",
        ),
    )

    quiet: bool = field(
        default=False,
        metadata=dict(
            title="Quiet",
            description="Skip printing non-error messages to stderr.",
        ),
    )

    color: bool = field(
        default=False,
        metadata=dict(
            title="Colorized diff",
            description="If --diff is selected showed diff is colorized.",
        ),
    )

    definitions: Iterable[Path] = field(
        default=tuple(),
        metadata=dict(
            title="Definitions",
            description=doc(
                """
    Files or directories containing custom command definitions (functions or macros).
    If only - is provided custom definitions, if there are any, are taken from stdin instead.
    Commands from not deprecated CMake native modules don't have to be provided.
    See: https://cmake.org/cmake/help/latest/manual/cmake-modules.7.html
                """
            ),
        ),
    )

    list_expansion: ListExpansion = field(
        default=ListExpansion.FavourInlining,
        metadata=dict(
            title="List expansion",
            description=doc(
                """
    Switch controls how code is expanded into multiple lines
    when it's not possible to keep it formatted in one line.
                """
            ),
        ),
    )

    workers: int = field(
        default=number_of_workers(),
        metadata=dict(
            title="Workers",
            description="Number of workers used to format multiple files in parallel.",
        ),
    )

    def summary(self):
        hasher = sha1()
        hasher.update(repr(self).encode("utf-8"))
        return hasher.hexdigest()


def make_configuration_file(configuration):
    d = asdict(configuration)
    d["list_expansion"] = d["list_expansion"].value
    result = f"""
# yaml-language-server: $schema=https://raw.githubusercontent.com/BlankSpruce/gersemi/master/gersemi/configuration.schema.json

{yaml.safe_dump(d)}
    """
    return result.strip()


def make_default_configuration_file():
    return make_configuration_file(Configuration())


def find_common_parent_path(paths: Iterable[Path]) -> Path:
    return Path(os.path.commonpath([path.absolute() for path in paths]))


def find_dot_gersemirc(paths: Iterable[Path]) -> Optional[Path]:
    lowest_level_common_parent = find_common_parent_path(paths)
    for parent in chain(
        [lowest_level_common_parent], lowest_level_common_parent.parents
    ):
        maybe_found = list(parent.glob(".gersemirc"))
        if maybe_found:
            return maybe_found[0]

    return None


@contextmanager
def enter_directory(target_directory):
    original = Path(".").resolve()
    try:
        os.chdir(target_directory)
        yield
    finally:
        os.chdir(original)


def normalize_definitions(definitions):
    return [Path(d).resolve() for d in definitions]


def sanitize_list_expansion(list_expansion):
    legal_values = [e.value for e in ListExpansion]
    if list_expansion in legal_values:
        return ListExpansion(list_expansion)
    raise RuntimeError(
        f"Unsupported list_expansion: '{list_expansion}'. Legal values: {', '.join(legal_values)}"
    )


def sanitize_workers(workers):
    return max(1, workers)


def load_configuration_from_file(configuration_file_path: Path) -> Configuration:
    with enter_directory(configuration_file_path.parent):
        with open(configuration_file_path, "r", encoding="utf-8") as f:
            config = yaml.safe_load(f.read())
            if "definitions" in config:
                config["definitions"] = normalize_definitions(config["definitions"])
            if "list_expansion" in config:
                config["list_expansion"] = sanitize_list_expansion(
                    config["list_expansion"]
                )
            if "workers" in config:
                config["workers"] = sanitize_workers(config["workers"])
            if "indent" in config:
                config["indent"] = indent_type(config["indent"])
        return Configuration(**config)


def override_configuration_with_args(
    configuration: Configuration, args
) -> Configuration:
    parameters = [field.name for field in fields(Configuration)]
    for param in parameters:
        value = getattr(args, param)
        if not value:
            continue
        if param == "definitions":
            value = normalize_definitions(value)
        if param == "list_expansion":
            value = sanitize_list_expansion(value)
        if param == "workers":
            value = sanitize_workers(value)
        setattr(configuration, param, value)
    return configuration


def make_configuration(args) -> Configuration:
    configuration_file_path = find_dot_gersemirc(args.sources)
    if configuration_file_path is not None:
        result = load_configuration_from_file(configuration_file_path)
    else:
        result = Configuration()

    return override_configuration_with_args(result, args)



from dataclasses import dataclass
from pathlib import Path
from gersemi.formatter import Formatter
from gersemi.utils import smart_open


@dataclass
class FormattedFile:
    before: str
    after: str
    newlines_style: str
    path: Path


def get_newlines_style(code: str) -> str:
    crlf = "\r\n"
    if crlf in code:
        return crlf

    return "\n"


def translate_newlines_to_line_feed(code: str) -> str:
    return code.replace("\r\n", "\n").replace("\r", "\n")


def format_file(path: Path, formatter: Formatter) -> FormattedFile:
    with smart_open(path, "r", newline="") as f:
        code = f.read()

    newlines_style = get_newlines_style(code)
    code = translate_newlines_to_line_feed(code)
    return FormattedFile(
        before=code,
        after=formatter.format(code),
        newlines_style=newlines_style,
        path=path,
    )



import pytest
from gersemi.custom_command_definition_finder import (
    find_custom_command_definitions,
    get_just_definitions,
)
from gersemi.dumper import Dumper


def test_custom_command_without_keyworded_arguments_formatting(
    parser_with_postprocessing,
):
    given = """function(some_custom_command_without_keyworded_arguments only positional arguments)
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(long__________________________________________________ positional__________________________________________________ arguments__________________________________________________)
"""

    expected = """function(
    some_custom_command_without_keyworded_arguments
    only
    positional
    arguments
)
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(
    long__________________________________________________
    positional__________________________________________________
    arguments__________________________________________________
)
"""

    parsed = parser_with_postprocessing.parse(given)
    definitions = get_just_definitions(find_custom_command_definitions(parsed))
    dumper = Dumper(width=80, indent_size=4, custom_command_definitions=definitions)

    formatted = dumper.visit(parsed)

    assert formatted == expected


def test_custom_command_without_keyworded_arguments_with_disabled_reformatting(
    parser_with_postprocessing,
):
    given = """function(some_custom_command_without_keyworded_arguments only positional arguments)
    # gersemi: ignore
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(long__________________________________________________ positional__________________________________________________ arguments__________________________________________________)
"""

    expected = """function(
    some_custom_command_without_keyworded_arguments
    only
    positional
    arguments
)
    # gersemi: ignore
    message(STATUS "some_custom_command_without_keyworded_arguments")
endfunction()

some_custom_command_without_keyworded_arguments(short positional arguments)

some_custom_command_without_keyworded_arguments(long__________________________________________________ positional__________________________________________________ arguments__________________________________________________)
"""

    parsed = parser_with_postprocessing.parse(given)
    definitions = get_just_definitions(find_custom_command_definitions(parsed))
    dumper = Dumper(width=80, indent_size=4, custom_command_definitions=definitions)

    formatted = dumper.visit(parsed)

    assert formatted == expected


@pytest.mark.parametrize(
    "given",
    [
        """function(some_custom_command_without_keyworded_arguments only positional arguments)
endfunction()
""",
        """function(some_custom_command_without_keyworded_arguments only positional arguments)

endfunction()
""",
    ],
)
def test_can_deal_with_empty_body_in_custom_command_definition(
    parser_with_postprocessing, given
):
    expected = """function(
    some_custom_command_without_keyworded_arguments
    only
    positional
    arguments
)
endfunction()
"""

    parsed = parser_with_postprocessing.parse(given)
    definitions = get_just_definitions(find_custom_command_definitions(parsed))
    dumper = Dumper(width=80, indent_size=4, custom_command_definitions=definitions)

    formatted = dumper.visit(parsed)

    assert formatted == expected



from typing import Callable
from lark import Tree, Token
from lark.visitors import Interpreter
from gersemi.types import Node


def is_tree(tree_type: str) -> Callable[[Node], bool]:
    return lambda element: isinstance(element, Tree) and element.data == tree_type


def is_token(token_type: str) -> Callable[[Node], bool]:
    return lambda element: isinstance(element, Token) and element.type == token_type


is_block = is_tree("block")
is_bracket_argument = is_tree("bracket_argument")
is_bracket_comment = is_tree("bracket_comment")
is_commented_argument = is_tree("commented_argument")
is_keyword_argument = is_tree("keyword_argument")
is_line_comment = is_tree("line_comment")
is_multi_value_argument = is_tree("multi_value_argument")
is_one_value_argument = is_tree("one_value_argument")
is_option_argument = is_tree("option_argument")
is_positional_arguments = is_tree("positional_arguments")
is_quoted_argument = is_tree("quoted_argument")
is_section = is_tree("section")
is_two_word_argument = is_tree("two_word_argument")
is_unquoted_argument = is_tree("unquoted_argument")

is_newline = is_token("NEWLINE")


def is_comment(element):
    return is_bracket_comment(element) or is_line_comment(element)


class ContainsLineComment(Interpreter):
    def __default__(self, _) -> bool:
        return False

    def line_comment(self, _) -> bool:
        return True

    def _visit(self, tree: Tree) -> bool:
        subtrees = filter(lambda node: isinstance(node, Tree), tree.children)
        return any(map(self.visit, subtrees))

    arguments = _visit
    commented_argument = _visit
    unary_operation = _visit
    binary_operation = _visit


def contains_line_comment(nodes) -> bool:
    visit = ContainsLineComment().visit
    return any(map(lambda node: isinstance(node, Tree) and visit(node), nodes))


def is_keyword(keyword):
    def impl(node):
        if is_unquoted_argument(node):
            return node.children[0] == keyword

        if is_quoted_argument(node):
            return (len(node.children) > 0) and (node.children[0] == keyword)

        if is_bracket_argument(node):
            return (len(node.children) > 0) and (node.children[0] == keyword)

        if is_commented_argument(node):
            return (len(node.children) > 0) and impl(node.children[0])

        return False

    return impl


class KeywordMatcher:
    def __init__(self, keywords):
        self.keywords = keywords

    def __call__(self, other):
        for k in self.keywords:
            if isinstance(k, str):
                if is_keyword(k)(other):
                    return True

            if isinstance(k, tuple) and is_keyword_argument(other):
                front, *_, back = other.children
                front_pattern, back_pattern = k
                if is_keyword(front_pattern)(front) and is_keyword(back_pattern)(back):
                    return True

        return False


def is_one_of_keywords(keywords):
    return KeywordMatcher(keywords)



from contextlib import contextmanager
from typing import Iterable, Mapping
from lark import Tree
from gersemi.ast_helpers import (
    is_multi_value_argument,
    is_one_of_keywords,
    is_positional_arguments,
)
from gersemi.keywords import KeywordMatcher
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


Sections = Mapping[KeywordMatcher, Mapping[KeywordMatcher, Iterable[KeywordMatcher]]]


def create_section_patch(section, old_class):
    def get(key):
        return section.get(key, [])

    class Impl(old_class):
        options = get("options")
        one_value_keywords = get("one_value_keywords")
        multi_value_keywords = get("multi_value_keywords")

    return Impl


class SectionAwareCommandInvocationDumper(ArgumentAwareCommandInvocationDumper):
    sections: Sections = {}

    @contextmanager
    def _update_section_characteristics(self, keyword):
        old_class = type(self)
        try:
            self.__class__ = create_section_patch(self.sections[keyword], old_class)
            yield
        finally:
            self.__class__ = old_class

    def _get_matcher(self, keyword):
        for item in self.sections:
            matcher = is_one_of_keywords([item])
            if matcher(keyword):
                return item
        return None

    def _split_multi_value_argument(self, tree):
        front_node, *rest = tree.children

        keyword_matcher = self._get_matcher(front_node)
        if keyword_matcher is None:
            return tree

        with self._update_section_characteristics(keyword_matcher):
            subarguments = [front_node]
            for arg in self._split_arguments(rest):
                if not is_positional_arguments(arg):
                    subarguments.append(arg)
                    continue

                subarguments.extend(arg.children)

            return Tree("section", subarguments)

    def _split_arguments(self, arguments):
        preprocessed = super()._split_arguments(arguments)
        return [
            (
                self._split_multi_value_argument(child)
                if is_multi_value_argument(child)
                else child
            )
            for child in preprocessed
        ]

    def section(self, tree):
        result = self._try_to_format_into_single_line(tree.children, separator=" ")
        if result is not None:
            return result

        header, *rest = tree.children
        begin = self.visit(header)
        if len(rest) == 0:
            return begin

        with self.indented():
            formatted_values = "\n".join(map(self.visit, rest))
        return f"{begin}\n{formatted_values}"



import os
from setuptools import setup, find_packages


HERE = os.path.dirname(os.path.realpath(__file__))

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

about = {}
with open(os.path.join(HERE, "gersemi", "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), about)  # pylint: disable=exec-used

setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=about["__url__"],
    packages=find_packages(include=["gersemi", "gersemi.*"]),
    package_data={"gersemi": ["cmake.lark", "builtin_commands"]},
    install_requires=[
        "appdirs",
        "dataclasses",
        "lark>=1.0",
        "pyyaml>=5,<7",
    ],
    extras_requires=["colorama>=0.4"],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["gersemi = gersemi.__main__:main"]},
    license=about["__license__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Quality Assurance",
    ],
)



from gersemi.custom_command_definition_finder import (
    find_custom_command_definitions,
    get_just_definitions,
)
from gersemi.dumper import Dumper
from .tests_generator import generate_input_only_tests


custom_command_to_format = """
seven_samurai(
three standalone arguments
KAMBEI KATSUSHIRO
GOROBEI foo HEIHACHI bar KYUZO baz
SHICHIROJI foo bar baz KIKUCHIYO bar baz foo)
"""

custom_command_properly_formatted = """Seven_Samurai(
    three
    standalone
    arguments
    KAMBEI
    KATSUSHIRO
    GOROBEI foo
    HEIHACHI bar
    KYUZO baz
    SHICHIROJI foo bar baz
    KIKUCHIYO bar baz foo
)
"""


def create_dumper(custom_command_definitions):
    return Dumper(
        width=80, indent_size=4, custom_command_definitions=custom_command_definitions
    )


def test_custom_command_generated_dumper(
    parser_with_postprocessing, case
):  # pylint: disable=redefined-outer-name
    parsed_function_def = parser_with_postprocessing.parse(case.content)
    parsed_function = parser_with_postprocessing.parse(custom_command_to_format)

    definitions = get_just_definitions(
        find_custom_command_definitions(parsed_function_def)
    )
    dumper = create_dumper(definitions)

    custom_command_formatted = dumper.visit(parsed_function)

    assert custom_command_formatted == custom_command_properly_formatted


pytest_generate_tests = generate_input_only_tests(
    where="custom_command_formatting", input_extension=".cmake"
)
